(*  Converted from Haskell to CaML (Sun Aug 16 13:05:13 EST 1998) 
\chapter{Lookup table}

This module {\em really} could do with some performance improvements.
Better yet,  I should take something out of the Chalmer's Haskell compiler's
standard library.


*)
module LUtable where


data 
   LUtable key value
        =  
           Node key value (LUtable key value) (LUtable key value)
           | NonNode
               deriving Show

data LU_returnable item = Success item | Failure

use_ret :: (LU_returnable item) -> item
use_ret Failure = error "Failure"
use_ret (Success x) = x


new_LUtable :: Ord key => LUtable key value
new_LUtable = NonNode

isempty_LUtable :: (LUtable key value) -> Bool
isempty_LUtable NonNode = True
isempty_LUtable _ = False

minimumEntry :: (LUtable key value) -> (key,value)
minimumEntry (Node k v NonNode _) = (k,v)
minimumEntry (Node _ _ b _) = minimumEntry b
minimumEntry (NonNode) = error "No minimum entry for empty LUtable"

maximumEntry :: (LUtable key value) -> (key,value)
maximumEntry (Node k v _ NonNode) = (k,v)
maximumEntry (Node _ _ _ a) = maximumEntry a
maximumEntry (NonNode) = error "No maximum entry for empty LUtable"

insert :: Ord key => (LUtable key value) -> (key,value) -> (LUtable key value)
insert NonNode (k,v) = Node k v NonNode NonNode
insert (Node k v before after) (k_alt,v_alt)
 | k_alt < k = Node k v (insert before (k_alt,v_alt)) after
 | k_alt > k = Node k v before (insert after (k_alt,v_alt))
 | otherwise = Node k v_alt before after
 

remove :: Ord key => (LUtable key value) -> key 
          ->  LU_returnable (LUtable key value)
remove NonNode _ = Failure
remove (Node k v before after) k_alt
 | k_alt < k = Success (Node k v (use_ret (remove before k_alt)) after)
 | k_alt > k = Success (Node k v before (use_ret (remove after k_alt)))
 | not (isempty_LUtable before)   = 
       Success (Node beforekey beforevalue (use_ret (remove before beforekey)) after)
 | not (isempty_LUtable after)   = 
             Success (Node afterkey aftervalue before (use_ret (remove after afterkey)))
 | otherwise = Success (NonNode)
     where (beforekey,beforevalue) = maximumEntry before
           (afterkey,aftervalue) = minimumEntry after
 
 
removeMany :: Ord key => (LUtable key value) -> [key] -> 
                  LU_returnable (LUtable key value)
removeMany l [] = Success l
removeMany l (x:xs) = cnd (remove l x)
  where cnd Failure = Failure
        cnd (Success s) = removeMany s xs
        
insertList :: Ord key => (LUtable key value) -> [(key,value)] 
                           -> (LUtable key value)
insertList l [] = l
insertList l (obj:objs) = insertList (insert l obj) objs

valueOf :: Ord key => (LUtable key value) -> key -> LU_returnable value
valueOf NonNode _ = Failure
valueOf (Node k v before after) k_alt 
  | k > k_alt  = valueOf before k_alt
  | k < k_alt  = valueOf after k_alt
  | otherwise  = Success v

predecessors :: Ord key => (LUtable key value) -> key -> (LUtable key value)
predecessors NonNode _ = NonNode
predecessors (Node k v n1 n2) k_alt 
  | k_alt < k  = predecessors n1 k_alt
  | k_alt > k = Node k v n1 (predecessors n2 k_alt)
  | otherwise = n1
  
postecessors :: Ord key => (LUtable key value) -> key -> (LUtable key value)
postecessors NonNode _ = NonNode
postecessors (Node k v n1 n2) k_alt
 | k < k_alt = Node k v (postecessors n1 k_alt) n2 
 | k > k_alt = postecessors n2 k_alt
 | otherwise = n2
 
mergeTables :: (Eq key,Ord key) => (LUtable key value) -> (LUtable key value)  ->
                (key -> value -> value -> value) -> (LUtable key value)    	
mergeTables NonNode NonNode _ = NonNode
mergeTables NonNode n       _ = n
mergeTables n       NonNode _ = n
mergeTables (Node k v n1 n2) n@(Node k_alt v_alt n1_alt n2_alt) merge_func
  | k == k_alt
          = Node k (merge_func k v v_alt)
                             (mergeTables n1 n1_alt merge_func)
                             (mergeTables n2 n2_alt merge_func)
  | otherwise = insert
                  (Node k v
                     (mergeTables n1 (predecessors n k) merge_func)
                     (mergeTables n2 (postecessors n k) merge_func)
                  ) (k_alt,v_alt)

mergeMany :: Ord key => [LUtable key value] -> (key -> value -> value -> value) 
                                                 -> (LUtable key value)
mergeMany [] _ = new_LUtable
mergeMany (x:xs) f = mergeTables x (mergeMany xs f) f

toList :: (LUtable key value) -> [(key,value)]
toList NonNode = []
toList (Node k v n1 n2) = ((k,v) : toList n1) ++ toList n2

toSortedList :: (LUtable key value) -> [(key,value)]
toSortedList NonNode = []
toSortedList (Node k v n1 n2) =  
  (toSortedList n1) ++ [(k,v)] ++ (toSortedList n2)

equal :: (Eq key,Eq value) => (LUtable key value) -> (LUtable key value) -> Bool
equal table1 table2 = (toSortedList table1) == (toSortedList table2)

instance (Eq key,Eq value) => Eq (LUtable key value)
 where
  table1 == table2 = toSortedList table1 == toSortedList table2

instance (Ord key,Ord value) => Ord (LUtable key value)
 where
   table1 <= table2 = toSortedList table1 < toSortedList table2

(**)
