(*  Converted from Haskell to CaML (Sun Aug 16 13:05:13 EST 1998) 
\chapter{Gate Definitions}

This is a rather boring file that contains definitions for all the
gates defined in Qgol.  The only interesting functions here are
\begin{description}
\item[Gate] The definition of the Gate datatype
\item[numberOfInputs] How many inputs a gate has
\item[numberOfOutputs]  How many outputs a gate has
\item[gate\_operation] What function should we apply to a ProcessorState for a given gate?
\item[strings2gate] For reading in a collection of gates from a text file
\end{description}

I think we could chop out a lot of unnecessary code by turning
\texttt{strings2gate} into a \texttt{Read} instance.  This might change
the file format a little,  but that hardly seems a problem.  (The original
file format was chosen to make it simple to type in by hand!)

\section{Preliminaries}
*)
open Complex
open Qgol_type

(*

\section{Gate Datatype}
*)
type gate = 
    PlusGate | MinusGate | TimesGate | Quotient | Div | Mod | Power | XAmodN
  | Cfrac | Unfrac | Entuple | Untuple | Enlist | Unlist
  | Split | Unsplit | Zero | UnZero | Positive
  | UnPositive | Toffoli | Divincenzo of float | Fourier
  | Not | Smolin of float | Deutsch of float | Source of Qgol_type.t
  | Observe of int | IsMonad | UnIsMonad | IsReal | UnIsReal
  | IsComplex | UnIsComplex | IsTuple | UnIsTuple
  | IsInteger | UnIsInteger | IsList | UnIsList
  | IsEmpty | UnIsEmpty  ;;

(*       deriving (Eq,Show)  ... harder *)
(*

\section{The important functions}
*)
let numberOfInputs g = match g with
 XAmodN -> 3
| PlusGate |  MinusGate |  TimesGate |  Quotient  |  Div  |  Mod  | Power -> 2
|  Cfrac |  Entuple |  Enlist -> 2
|  Unfrac |  Untuple |  Unlist -> 1
|  Split |  Zero |  Positive -> 1
|  Unsplit |  UnZero |  UnPositive -> 2

|  Toffoli -> 3
|  (Divincenzo _) -> 2
|  Fourier -> 2
|  Not -> 1
|  (Smolin _) -> 2
|  (Deutsch _) -> 3
|  (Source _) -> 0
|  (Observe _) -> 1
|  IsMonad | IsReal | IsComplex | IsTuple | IsInteger | IsList | IsEmpty -> 1
| UnIsMonad |UnIsReal |UnIsComplex |UnIsTuple |UnIsInteger |UnIsList|UnIsEmpty -> 2


let numberOfOutputs g = match g with
   PlusGate |  MinusGate |  TimesGate |  Quotient |  Div |  Mod  |  Power -> 2
 |  XAmodN -> 4
 |  Cfrac |  Entuple |  Enlist -> 1
 |  Unfrac |  Untuple |  Unlist -> 2
 |  Split |  Zero |  Positive -> 2
 |  Unsplit |  UnZero |  UnPositive -> 1
 |  Toffoli -> 3
 |  (Divincenzo _) -> 2
 |  Fourier -> 2
 |  Not -> 1
 |  (Smolin _) -> 2
 |  (Deutsch _) -> 3
 |  (Source _) -> 1
 |  (Observe _) -> 0
 | IsMonad | IsReal | IsComplex | IsTuple | IsInteger | IsList | IsEmpty -> 2
 |UnIsMonad|UnIsReal|UnIsComplex|UnIsTuple|UnIsInteger|UnIsList|UnIsEmpty -> 1

(*

This next stuff I think would be better implemented as some kind of
lexer or grammar or something,  but I haven't got around to it yet.

*)
let strings2gate str1 str2 = match str1 with
| "plus" -> PlusGate | "minus" -> MinusGate | "times" -> TimesGate
| "quotient" -> Quotient | "div" -> Div | "mod" -> Mod
| "power" -> Power | "xamodn" -> XAmodN
| "cfrac" -> Cfrac| "unfrac" -> Unfrac
| "entuple" -> Entuple| "untuple" -> Untuple
| "enlist" -> Enlist| "unlist" -> Unlist
| "split" -> Split| "unsplit" -> Unsplit
| "zero" -> Zero| "un-zero" -> UnZero
| "positive" -> Positive| "un-positive" -> UnPositive
| "toffoli" -> Toffoli
| "divincenzo" -> Divincenzo (float_of_string str2)
| "fourier" -> Fourier
| "not" -> Not
| "smolin" -> Smolin (float_of_string str2)
| "deutsch" -> Deutsch (float_of_string str2)
| "source" when str2 = "." -> Source (Qgol_type.Monad)
| "source" when str2 = "[]" -> Source (Qgol_type.List [])
| "source" when str2 = "true" -> Source (Qgol_type.Boolean true)
| "source" when str2 = "false" -> Source (Qgol_type.Boolean false)
| "source" when Recognize_numbers.is_integer str2 
         -> Source (Intger (int_of_string str2))
| "source" when Recognize_numbers.is_real str2     
         -> Source (Dbl (float_of_string str2))
| "source"                    -> Source (Str str2)
| "is_monad" -> IsMonad
| "un-is_monad" -> UnIsMonad
| "is_real" -> IsReal
| "un-is_real" -> UnIsReal
| "is_complex" -> IsComplex
| "un-is_complex" -> UnIsComplex
| "is_tuple" -> IsTuple
| "un-is_tuple" -> UnIsTuple
| "is_list" -> IsList
| "un-is_list" -> UnIsList
| "is_empty" -> IsEmpty
| "un-is_empty" -> UnIsEmpty
| "observe" -> Observe (int_of_string str2)
| _ -> (prerr_endline ("Couldn't gatify \"" ^ str1 ^ ":" ^ str2 ^ "\""); exit 1)
;;

(*

\label{gate operation definitions}

We are about to enter into a long section where we define
functions for each Gate.  Since we haven't (and can't have,  I suspect)
existential typing,  each function will run-time check its arguments.
If the run time check fails,  we'll use \texttt{wrong\_function\_usage}.

I think that \texttt{print\_arguments} could be turned simply into a
Shows instance.  (Or that something like it should be done).

*)

let print_arguments args = Util.string_join "; " (List.map Qgol_type.show args);;

let wrong_function_usage fun_name args
 = prerr_endline (fun_name ^ " called with wrong arguments: " ^
          (print_arguments args)) ; exit 1;;


let error str = prerr_endline str ; exit 1;;
(*

\section{Classical Functions}

*)
let funcPlus u = match u with
  [a;b] -> [a;(a%+%b)]
| _ -> wrong_function_usage "funcPlus" u;;

let funcMinus u = match u with
 [a;b] -> [a;(a%-%b)]
| _ -> wrong_function_usage "funcMinus" u;;

let funcTimes u = match u with
  [a;b] -> [a;(a%*%b)]
| _ -> wrong_function_usage "funcTimes" u;;

let funcQuotient u = match u with
  [Dbl a;Dbl b] -> [Dbl a;Dbl (a/.b)]
|  _ -> wrong_function_usage "Quotient" u;;

let funcDiv u = match u with
  [Intger a;Intger b] -> [Intger (a / b);Pair (Intger a,Intger b)]
| _ -> wrong_function_usage "funcDiv" u;;

let funcMod u = match u with
  [Intger a;Intger b] -> [Intger (a mod b);Pair (Intger a,Intger b)]
| _ -> wrong_function_usage "funcMod" u;;

let funcPower u = prerr_endline "funcPower not implemented yet" ; exit 1 ;;

let funcXAmodN u = match u with
 [Intger x;Intger a; Intger n]
 -> [Intger x; Intger a; Intger n; Intger (Misc_funcs.moduloPower (x,a,n))]
| _ -> wrong_function_usage "funcXAmodN" u;;

(*
let funcCfrac u = match u with
let funcCfrac [Intger a; Intger b] -> 
    [List (map (\x -> Intger x) (toList (rationalToCfrac (a % b))))]
let funcCfrac u -> wrong_function_usage "funcCfrac" u;;

let funcUnfrac u = match u with
let funcUnfrac [List a] -> [Intger (numerator r),Intger (denominator r)]
  where r = cfracToRational (Continued_Fractions.Continued_Fraction list)
        list = map (\x -> case (x) of
                       (Intger y) -> y
                       (_) -> error "Not an integer in Unfracing"
                     ) a
let funcUnfrac u -> wrong_function_usage "funcUnfrac" u;;
*)

let funcEnPair u = match u with
 [a;b] -> [Pair (a,b)]
| _ -> wrong_function_usage "funcEnPair" u;;

let funcUnPair u = match u with
 [Pair (a,b)] -> [a;b]
| _ -> wrong_function_usage "funcUnPair" u;;

let funcEnlist u = match u with
 [a;List b] -> [List (a::b)]
| _ -> wrong_function_usage "funcEnlist" u;;

let funcUnlist u = match u with
 [List (a::b)] -> [a;List b]
| _ -> wrong_function_usage "funcUnlist" u;;

let funcSplit u = match u with
 [a] -> [a;a]
| _ -> wrong_function_usage "funcSplit" u;;

let funcUnsplit u = match u with
 [a;b] when a=b -> [a]
| [a;b] -> (prerr_endline ("attempt to unify non-equal values" ^ print_arguments [a;b]) ; exit 1)
| _ -> wrong_function_usage "funcUnsplit" u;;

(*

\section{Observe}
*)
let funcObserve x = prerr_endline "funcObserve isn't this simple" ; exit 1;;
(*

\section{Is-style functions}
*)
let funcIsMonad x = match x with
 Monad -> true
|  _ -> false;;

let funcIsReal x = match x with
 (Dbl  _) -> true
| _ -> false;;

let funcIsComplex x = match x with
(* funcIsComplex (Cmplx _) -> true
   in the past,  Type had an entry Cmplx *)
 _ -> false;;

let funcIsList x = match x with
 (List _) -> true
|  _ -> false;;

let funcIsInteger x = match x with
 (Intger _) -> true
| _ -> false;;

let funcIsPair x = match x with
 (Pair (_,_)) -> true
| _ -> false;;

let funcIsEmpty x = match x with
 (List []) -> true
| _ -> false;;

let funcIsZero x = match x with
  (Intger 0) -> true
|  (Dbl 0.0) -> true
| _ -> false;;

let funcIsPositive u = match u with
 (Intger x) -> x > 0
| (Dbl x) -> x > 0.0
| _ -> false;;



(*

\section{Quantum functions}
*)
let funcToffoli x = match x with
[Qgol_type.Boolean a;Qgol_type.Boolean b;Qgol_type.Boolean c] when  a && b ->
       [([Qgol_type.Boolean a;Qgol_type.Boolean b;Boolean (not c)],unity)]
| [Boolean a;Boolean b;Boolean c] -> [([Boolean a;Boolean b;Boolean c],unity)]
|  _ -> error "funcToffoli with wrong args"

let funcDivincenzo phi args = match args with
  [Boolean false;Boolean a] ->  [([Boolean false;Boolean a],unity)]
| [Boolean true;Boolean false] ->
     [([Boolean true;Boolean false],{re=cos phi; im = 0.0});
      ([Boolean true;Boolean true],{re=sin phi; im = 0.0})]
| [Boolean true;Boolean true] ->
     [([Boolean true;Boolean false],{re=sin (~-.phi);im=0.0});
      ([Boolean true;Boolean true],{re=cos phi;im=0.0})]
| _ -> error "funcDivincenzo with wrong args"

let funcFourier args = match args with
 [Intger size;Intger smear_me]
 -> List.map (function (x,y)->([Intger size;Intger y],x)) (Misc_funcs.fourierTransform size smear_me)
|  _ -> wrong_function_usage "funcFourier" args;;

(*

\texttt{funcNot} is for Deutsch's square-root of \textsf{NOT} function\cite{Deutsch}.
*)
let funcNot args = match args with
 [Boolean true] ->
    [([Boolean true],{re = (1.0/.sqrt(2.0));im=0.0});
     ([Boolean false],{re=(1.0/.sqrt(2.0));im=0.0})]
| [Boolean false] ->
    [([Boolean true],{re = (-1.0/.sqrt(2.0));im = 0.0});
     ([Boolean false],{re = (1.0/.sqrt(2.0));im =0.0})]
|  _ -> error "funcNot with wrong args";;

let funcSmolin phi args = match args with
 [Boolean false; Boolean false] ->
      [([Boolean false;Boolean false],e_to_the_i phi)]
| [Boolean x; Boolean y] -> [([Boolean x;Boolean y], unity)]
| _ -> error "funcSmolin with wrong args";;

let funcDeutsch theta args = match args with
_ -> error "funcDeutsch not implemented yet";;

let funcSource t args = match args with
 [] -> [([t],unity)]
| _ -> error "funcSource with wrong arg count";;
(*

\section{Function definitions}

*)

let fromClassical func = (function x -> [(func x,unity)]);;

let fromIsToIs func args = match args with
 [x] -> [([x;Boolean (func x)],unity)]
| _ -> error "wrong number of arguments for fromIs";;

let fromIsToUn func args = match args with
  [value;Boolean condition] when condition && (func value) -> [([value],unity)]
| [_;Boolean _] -> error "reversibility attempt failed"
| _ -> error "fromIsToUn with incorrect argument count / style";;



let gate_operation g = match g with
 PlusGate -> fromClassical funcPlus
| MinusGate -> fromClassical funcMinus
| TimesGate -> fromClassical funcTimes
| Quotient -> fromClassical funcQuotient
| Div -> fromClassical funcDiv
| Mod -> fromClassical funcMod
| Power -> fromClassical funcPower
| XAmodN -> fromClassical funcXAmodN
(*| Cfrac -> fromClassical funcCfrac
| Unfrac -> fromClassical funcUnfrac*)
| Entuple -> fromClassical funcEnPair
| Untuple -> fromClassical funcUnPair
| Enlist -> fromClassical funcEnlist
| Unlist -> fromClassical funcUnlist
| Split -> fromClassical funcSplit
| Unsplit -> fromClassical funcUnsplit
| Toffoli -> funcToffoli
| (Divincenzo x) -> funcDivincenzo x
| Fourier -> funcFourier
| Not -> funcNot
| (Smolin x) -> funcSmolin x
| (Deutsch x) -> funcDeutsch x
| (Source x) -> funcSource x
| (Observe x) -> funcObserve
| IsMonad -> fromIsToIs funcIsMonad
| UnIsMonad -> fromIsToUn funcIsMonad
| IsReal -> fromIsToIs funcIsReal
| UnIsReal -> fromIsToUn funcIsReal
| IsComplex -> fromIsToIs funcIsComplex
| UnIsComplex -> fromIsToUn funcIsComplex
| IsTuple -> fromIsToIs funcIsPair
| UnIsTuple -> fromIsToUn funcIsPair
| IsInteger -> fromIsToIs funcIsInteger
| UnIsInteger -> fromIsToUn funcIsInteger
| IsList -> fromIsToIs funcIsList
| UnIsList -> fromIsToUn funcIsList
| IsEmpty -> fromIsToIs funcIsEmpty
| UnIsEmpty -> fromIsToUn funcIsEmpty
| Zero -> fromIsToIs funcIsZero
| UnZero -> fromIsToUn funcIsZero
| Positive -> fromIsToIs funcIsPositive
| UnPositive -> fromIsToUn funcIsPositive
;;
(**)
