(*  Converted from Haskell to CaML (Sun Aug 16 13:05:14 EST 1998) 
    misc_funcs.ml
*)
open Complex
(*

\texttt{raise\_around} picks on entry out of the following matrix \ldots 
\[
\frac{1}{\sqrt{N}}
\left(
\begin{array}{ccccc}
1 &          1        &             1      &             1  & \ldots \\
1 & e^{\left(\frac{2 \pi i}{N}\right)} & e^{{\left(\frac{2 \pi i}{N}\right)}^2} &  e^{{\left(\frac{2 \pi i}{N}\right)}^3} & \ldots \\
1 & e^{{\left(\frac{2 \pi i}{N}\right)}^2} & e^{{\left(\frac{2 \pi i}{N}\right)}^4} &  e^{{\left(\frac{2 \pi i}{N}\right)}^6} & \ldots \\
1 & e^{{\left(\frac{2 \pi i}{N}\right)}^3} & e^{{\left(\frac{2 \pi i}{N}\right)}^6} &  e^{{\left(\frac{2 \pi i}{N}\right)}^9} & \ldots \\
\vdots & \vdots & \vdots & \vdots & \ddots
\end{array}
\right)
\]

\begin{description}
\item[size] The array is square,  size gives number of rows and columns.  Size is abbreviated to \texttt{N} in the above matrix
\item[row] Which row to return
\item[column] Which column
\end{description}

*)
let raise_around size row column = 
 let sizef = float size
 and rowf = float row
 and columnf = float column in
 let iteration_segment = 2.0 *. (Util.pi /. sizef)
 and one_on_root_N = {re = 1.0 /. (sqrt sizef) ; im = 0.0 }
 in
  (e_to_the_i (rowf *. columnf *. iteration_segment)) &*& one_on_root_N
;;
        

(*

\section{Fourier Transform}

*)
let fourierTransform size row =
 let column_numbers = Util.list_range 0 (size-1) in
   List.combine (List.map (raise_around size row) column_numbers) column_numbers;;



(*
 
\section{ModuloPower}

\texttt{ModuloPower} calculates $x^a \quad{}\textsf{mod}\quad{} n$
*)

let rec moduloPower (x,a,n) = 
 if (a < 0) then (prerr_endline "moduloPower called with a < 0" ; exit 1) else
 if (x < 0) then (prerr_endline "moduloPower called with x < 0" ; exit 1) else
 if (n < 0) then (prerr_endline "moduloPower called with n < 0" ; exit 1) else
 if (a = 0) then 1 else
 if (a = 1) then x mod n else
  let recursive_part  = moduloPower (x,(a / 2),n) in let
      squared_recursive_part = (recursive_part * recursive_part) mod n
  in 
    if (a mod 2 = 1) then ((x * squared_recursive_part) mod n)
    else squared_recursive_part
