(*  Converted from Haskell to CaML (Sun Aug 16 13:05:13 EST 1998) 
\chapter{Gate Definitions}

This is a rather boring file that contains definitions for all the
gates defined in Qgol.  The only interesting functions here are
\begin{description}
\item[Gate] The definition of the Gate datatype
\item[numberOfInputs] How many inputs a gate has
\item[numberOfOutputs]  How many outputs a gate has
\item[gate\_operation] What function should we apply to a ProcessorState for a given gate?
\item[strings2gate] For reading in a collection of gates from a text file
\end{description}

I think we could chop out a lot of unnecessary code by turning
\texttt{strings2gate} into a \texttt{Read} instance.  This might change
the file format a little,  but that hardly seems a problem.  (The original
file format was chosen to make it simple to type in by hand!)

\section{Preliminaries}
*)
open Complex

(*

\section{Gate Datatype}
*)
type gate = 
    PlusGate | MinusGate | TimesGate | Quotient | Div | Mod | Power | XAmodN
  | Cfrac | Unfrac | Entuple | Untuple | Enlist | Unlist
  | Split | Unsplit | Zero | UnZero | Positive
  | UnPositive | Toffoli | Divincenzo Double | Fourier
  | Not | Smolin Double | Deutsch Double | Source Type
  | Observe Integer | IsMonad | UnIsMonad | IsReal | UnIsReal
  | IsComplex | UnIsComplex | IsTuple | UnIsTuple
  | IsInteger | UnIsInteger | IsList | UnIsList
  | IsEmpty | UnIsEmpty  ;;

(*       deriving (Eq,Show)  ... harder *)
(*

\section{The important functions}
*)
numberOfInputs g = match g with
 XAmodN -> 3
| PlusGate |  MinusGate |  TimesGate |  Quotient  |  Div  |  Mod  | Power -> 2
|  Cfrac |  Entuple |  Enlist -> 2
|  Unfrac |  Untuple |  Unlist -> 1
|  Split |  Zero |  Positive -> 1
|  Unsplit |  UnZero |  UnPositive -> 2

|  Toffoli -> 3
|  (Divincenzo _) -> 2
|  Fourier -> 2
|  Not -> 1
|  (Smolin _) -> 2
|  (Deutsch _) -> 3
|  (Source _) -> 0
|  (Observe _) -> 1
|  IsMonad | IsReal | IsComplex | IsTuple | IsInteger | IsList | IsEmpty -> 1
| UnIsMonad |UnIsReal |UnIsComplex |UnIsTuple |UnIsInteger |UnIsList|UnIsEmpty -> 2


numberOfOutputs g = match g with
   PlusGate |  MinusGate |  TimesGate |  Quotient |  Div |  Mod  |  Power -> 2
 |  XAmodN -> 4
 |  Cfrac |  Entuple |  Enlist -> 1
 |  Unfrac |  Untuple |  Unlist -> 2
 |  Split |  Zero |  Positive -> 2
 |  Unsplit |  UnZero |  UnPositive -> 1
 |  Toffoli -> 3
 |  (Divincenzo _) -> 2
 |  Fourier -> 2
 |  Not -> 1
 |  (Smolin _) -> 2
 |  (Deutsch _) -> 3
 |  (Source _) -> 1
 |  (Observe _) -> 0
 | IsMonad | IsReal | IsComplex | IsTuple | IsInteger | IsList | IsEmpty -> 2
 |UnIsMonad|UnIsReal|UnIsComplex|UnIsTuple|UnIsInteger|UnIsList|UnIsEmpty -> 1

(*

This next stuff I think would be better implemented as some kind of
lexer or grammar or something,  but I haven't got around to it yet.

*)
strings2gate str1 str2 = match (str1,str2) with
| "plus" -> PlusGate | "minus" -> MinusGate | "times" -> TimesGate
| "quotient" -> Quotient | "div" -> Div | "mod" -> Mod
| "power" -> Power | "xamodn" -> XAmodN
| "cfrac" -> Cfrac| "unfrac" -> Unfrac
| "entuple" -> Entuple| "untuple" -> Untuple
| "enlist" -> Enlist| "unlist" -> Unlist
| "split" -> Split| "unsplit" -> Unsplit
| "zero" -> Zero| "un-zero" -> UnZero
| "positive" -> Positive| "un-positive" -> UnPositive
| "toffoli" -> Toffoli
| "divincenzo" -> Divincenzo (float_of_string str2)
| "fourier" -> Fourier
| "not" -> Not
| "smolin" -> Smolin (float_of_string str2)
| "deutsch" -> Deutsch (float_of_string str2)
| "source" when str2 = "." -> Source Monad
| "source" when str2 = "[]" -> Source (List [])
| "source" when str2 = "true" -> Source (Boolean True)
| "source" when str2 = "false" -> Source (Boolean False)
| "source" when is_integer str2  -> Source (Intger (int_of_string str2))
| "source" when is_real str2     -> Source (Dbl (float_of_string str2))
| "source"                    -> Source (Str str2)
| "is_monad" -> IsMonad
| "un-is_monad" -> UnIsMonad
| "is_real" -> IsReal
| "un-is_real" -> UnIsReal
| "is_complex" -> IsComplex
| "un-is_complex" -> UnIsComplex
| "is_tuple" -> IsTuple
| "un-is_tuple" -> UnIsTuple
| "is_list" -> IsList
| "un-is_list" -> UnIsList
| "is_empty" -> IsEmpty
| "un-is_empty" -> UnIsEmpty
| "observe" -> Observe (int_of_string str2)
| _ -> (prerr_endline ("Couldn't gatify \"" ^ str1 ^ ":" ^ str2 ^ "\""); exit 1)
;;

(*

\label{gate operation definitions}

We are about to enter into a long section where we define
functions for each Gate.  Since we haven't (and can't have,  I suspect)
existential typing,  each function will run-time check its arguments.
If the run time check fails,  we'll use \texttt{wrong\_function\_usage}.

I think that \texttt{print\_arguments} could be turned simply into a
Shows instance.  (Or that something like it should be done).

*)
wrong_function_usage fun_name args
 = prerr_endline (fun_name ^ " called with wrong arguments: " ^
          print_arguments args) ; exit 1;;

print_arguments args = Util.string_join "; " (List.map Qgol_type.show) args

error str = prerr_endline str ; exit 1;;
(*

\section{Classical Functions}

*)
funcPlus u = match u with
  [a;b] -> [a,(a%+%b)]
| _ = wrong_function_usage "funcPlus" u;;

funcMinus u = match u with
 [a;b] -> [a,(a-b)]
| _ -> wrong_function_usage "funcMinus" u;;

funcTimes u = match u with
funcTimes [a;b] -> [a,(a*b)]
| _ -> wrong_function_usage "funcTimes" u;;

funcQuotient u = match u with
funcQuotient [Dbl a;Dbl b] -> [Dbl a,Dbl (a/b)]
|  _ -> wrong_function_usage "Quotient" u;;

funcDiv u = match u with
funcDiv [Intger a;Intger b] -> [Intger (a `div` b),Pair (Intger a) (Intger b)]
| _ -> wrong_function_usage "funcDiv" u;;

funcMod u = match u with
funcMod [Intger a;Intger b] -> [Intger (a `mod` b),Pair (Intger a) (Intger b)]
| _ -> wrong_function_usage "funcMod" u;;

funcPower u = prerr_endline "funcPower not implemented yet" ; exit 1 ;;

funcXAmodN u = match u with
 [Intger x;Intger a; Intger n]
 -> [Intger x, Intger a, Intger n, Intger (MiscFuncs.moduloPower (x,a,n))]
| _ -> wrong_function_usage "funcXAmodN" u;;

(*
funcCfrac u = match u with
funcCfrac [Intger a; Intger b] -> 
    [List (map (\x -> Intger x) (toList (rationalToCfrac (a % b))))]
funcCfrac u -> wrong_function_usage "funcCfrac" u;;

funcUnfrac u = match u with
funcUnfrac [List a] -> [Intger (numerator r),Intger (denominator r)]
  where r = cfracToRational (Continued_Fractions.Continued_Fraction list)
        list = map (\x -> case (x) of
                       (Intger y) -> y
                       (_) -> error "Not an integer in Unfracing"
                     ) a
funcUnfrac u -> wrong_function_usage "funcUnfrac" u;;
*)

funcEnPair u = match u with
 [a;b] -> [Pair a b]
| _ -> wrong_function_usage "funcEnPair" u;;

funcUnPair u = match u with
 [Pair (a,b)] -> [a,b]
| _ -> wrong_function_usage "funcUnPair" u;;

funcEnlist u = match u with
 [a;List b] -> [List (a:b)]
| _ -> wrong_function_usage "funcEnlist" u;;

funcUnlist u = match u with
 [List (a::b)] -> [a,List b]
| _ -> wrong_function_usage "funcUnlist" u;;

funcSplit u = match u with
 [a] -> [a;a]
| _ -> wrong_function_usage "funcSplit" u;;

funcUnsplit u = match u with
 [a;b] when a=b -> [a]
 [a;b]-> (prerr_endline ("attempt to unify non-equal values" ++  print_arguments [a;b]) ; exit 1)
| _ -> wrong_function_usage "funcUnsplit" u;;

(*

\section{Observe}
*)
funcObserve _ = prerr_endline "funcObserve isn't this simple" ; exit 1;;
(*

\section{Is-style functions}
*)
funcIsMonad x = match x with
 Monad -> True
|  _ -> False;;

funcIsReal x = match x with
 (Dbl  _) -> True
| _ -> False;;

funcIsComplex x = match x with
--funcIsComplex (Cmplx _) -> True
-- in the past,  Type had an entry Cmplx
 _ -> False;;

funcIsList x = match x with
 (List _) -> True
|  _ -> False;;

funcIsInteger x = match x with
 (Intger _) -> True
| _ -> False;;

funcIsPair x = match x with
 (Pair _ _) -> True
| _ -> False;;

funcIsEmpty x = match x with
 (List []) -> True
| _ -> False;;

funcIsZero x = match x with
 (Intger 0) -> True
 (Dbl 0) -> True
| _ -> False;;

funcIsPositive u = match u with
 (Intger x) -> x > 0
 (Dbl x) -> x > 0.0
| _ -> False;;



(*

\section{Quantum functions}
*)
funcToffoli x = match x with
[Boolean a;Boolean b;Boolean c] when  a && b ->
       [([Boolean a;Boolean b;Boolean (not c)],unity)]
| [Boolean a;Boolean b;Boolean c] -> [([Boolean a;Boolean b;Boolean c],unity})]
|  _ -> error "funcToffoli with wrong args"

funcDivincenzo phi args = match args with
  [Boolean False;Boolean a] ->  [([Boolean False;Boolean a],unit)]
| [Boolean True,Boolean False] ->
     [([Boolean True,Boolean False],{re=cos phi; im = 0.0});
      ([Boolean True,Boolean True],{re=sin phi; im = 0.0})]
| [Boolean True,Boolean True] ->
     [([Boolean True,Boolean False],{re=sin (~-.phi);im=0.0});
      ([Boolean True,Boolean True],{re=cos phi;im=0.0})]
| _ -> error "funcDivincenzo with wrong args"

funcFourier args = match args with
 [Intger size,Intger smear_me]
 -> List.map (function (x,y)->([Intger size,Intger y],x)) (MiscFuncs.fourierTransform size smear_me)
|  _ -> wrong_function_usage "funcFourier" args;;

(*

\texttt{funcNot} is for Deutsch's square-root of \textsf{NOT} function\cite{Deutsch}.
*)
funcNot args = match args with
 [Boolean True] ->
    [([Boolean True],{re = (1.0/.sqrt(2.0));im=0.0});
     ([Boolean False],{re=(1.0/sqrt(2.0));im=0.0})]
| [Boolean False] ->
    [([Boolean True],{re = (-1.0/.sqrt(2.0));im = 0.0}),
     ([Boolean False],{re = (1.0/sqrt(2.0));im =0.0})]
|  _ -> error "funcNot with wrong args";;

funcSmolin phi args = match args with
 [Boolean False; Boolean False] ->
      [([Boolean False;Boolean False],e_to_the_i phi)]
| _ [Boolean x, Boolean y] -> [([Boolean x,Boolean y], unity)]
| _ -> error "funcSmolin with wrong args";;

funcDeutsch theta args = match args with
_ -> error "funcDeutsch not implemented yet";;

funcSource t args = match args with
 [] -> [([t],unity)]
| _ -> error "funcSource with wrong arg count";;
(*

\section{Function definitions}

*)

fromClassical func = (function x -> [(func x,1)])

fromIsToIs func args = match args with
 [x] -> [([x,Boolean (func x)],unity)]
| _ -> error "wrong number of arguments for fromIs"

fromIsToUn func args = match args with
  [value,Boolean condition] when condition && (func value) -> [([value],unity)]
| [_,Boolean _] -> error "reversibility attempt failed"
| _ -> error "fromIsToUn with incorrect argument count / style"




gate_operation PlusGate = fromClassical funcPlus
gate_operation MinusGate = fromClassical funcMinus
gate_operation TimesGate = fromClassical funcTimes
gate_operation Quotient = fromClassical funcQuotient
gate_operation Div = fromClassical funcDiv
gate_operation Mod = fromClassical funcMod
gate_operation Power = fromClassical funcPower
gate_operation XAmodN = fromClassical funcXAmodN
gate_operation Cfrac = fromClassical funcCfrac
gate_operation Unfrac = fromClassical funcUnfrac
gate_operation Entuple = fromClassical funcEnPair
gate_operation Untuple = fromClassical funcUnPair
gate_operation Enlist = fromClassical funcEnlist
gate_operation Unlist = fromClassical funcUnlist
gate_operation Split = fromClassical funcSplit
gate_operation Unsplit = fromClassical funcUnsplit
gate_operation Toffoli = funcToffoli
gate_operation (Divincenzo x) = funcDivincenzo x
gate_operation Fourier = funcFourier
gate_operation Not = funcNot
gate_operation (Smolin x) = funcSmolin x
gate_operation (Deutsch x) = funcDeutsch x
gate_operation (Source x) = funcSource x
gate_operation (Observe _) = funcObserve
gate_operation IsMonad = fromIsToIs funcIsMonad
gate_operation UnIsMonad = fromIsToUn funcIsMonad
gate_operation IsReal = fromIsToIs funcIsReal
gate_operation UnIsReal = fromIsToUn funcIsReal
gate_operation IsComplex = fromIsToIs funcIsComplex
gate_operation UnIsComplex = fromIsToUn funcIsComplex
gate_operation IsTuple = fromIsToIs funcIsPair
gate_operation UnIsTuple = fromIsToUn funcIsPair
gate_operation IsInteger = fromIsToIs funcIsInteger
gate_operation UnIsInteger = fromIsToUn funcIsInteger
gate_operation IsList = fromIsToIs funcIsList
gate_operation UnIsList = fromIsToUn funcIsList
gate_operation IsEmpty = fromIsToIs funcIsEmpty
gate_operation UnIsEmpty = fromIsToUn funcIsEmpty
gate_operation Zero = fromIsToIs funcIsZero
gate_operation UnZero = fromIsToUn funcIsZero
gate_operation Positive = fromIsToIs funcIsPositive
gate_operation UnPositive = fromIsToUn funcIsPositive
(**)
