(*  Converted from Haskell to CaML (Sun Aug 16 13:05:13 EST 1998) 
\chapter{Gate Dictionary}

Comments pending. 

*)

type gateId = int;;
type wireId = int;;

type wireInstance =  
    {    
     sourceSock :  int,
     sourceGate :  gateId ,
     destSock   :  int ,
     destGate   :  gateId
    };;

let dudWireId : wireId = -1;;
let rec dudwirelist n = if n = 0 then [] else dudWireId::(dudwirelist (n-1));;
   
type gateInstance = { 
   nature : Gatedefs.Gate , 
   outs : wireId list, 
   inps : wireId list
 } ;;

type gateDictionary = (gateId * gateInstance) list;;
type wireDictionary = (wireId * wireInstance) list;;
type circuitDictionary = (gateDictionary,wireDictionary,wireId list)

let fresh_wireId_list = Util.list_range 1000 2000;;

let makeDictionary x = (makeDictionary_alt x, [],[1000..]);;

let insert dictionary (key,value) = 
  let non_matching_keys = Util.filter (function (x,y) -> x != key) dictionary
      in (key,value)::non_matching_keys;



let instantiate (gate,gateid) =
  (gateid,{nature=gate,outs=(dudwirelist Gatedefs.numberOfOutputs gate),
	    inps = (dudwirelist Gatedefs.numberOfInputs gate)});;

let makeDictionary_alt = List.map instantiate;;

let makeConnection (gates,wires,unused_wires) (sourceId,sourceSock) (destId,destSock) =
 let wire::unused_wires_alt = unused_wires in
 let wires_alt = makeWire wires wire (sourceId,sourceSock) (destId,destSock)
 in (setOutputSocket (setInputSocket gates (destId,destSock) wire)
         (sourceId,sourceSock) wire , wires_alt , unused_wires_alt)
;;

let makeWire dict id (src,srcSock) (dst,dstSock) = 
  insert dictionary (id, {sourceGate = src, 
			   sourceSock = srcSock, 
			   destGate = dst, 
			   destSock = dstSock
			 });;
	     
let error x = prerr_endline x ; exit 1;;
       
let setOutputSocket dict (source,sourceSock) w_id =
 let old_instance = List.assoc source dict in
 let new_instance = { inps = old_instance.inps;
                      nature = old_instance.nature;
		      outs =  Smlcompat.listupdate (old_instance.outs) (sourceSock-1) w_id } in
 if (sourceSock > (numberOfOutputs old_instance.nature)) then
   (prerr_endline ("setOutputSocket: Request for too great a socket number of "
		   ^ (string_of_int sourceSock) ; exit 1) )
 else if (sourceSock < 0) = error "Request for negative socket number"
 else insert dict (source,new_instance);;
 
 
let setInputSocket dict (dest,destSock) w_id =
  let old_instance = use_ret (valueOf dict dest) in
  let new_instance = { nature = old_instance.nature;
		       outs = old_instance.outs;
		       inps = Smlcompat.listupdate (old_instance.inps) (destSock-1) w_id
		     }
  in if (destSock > numberOfInputs (nature old_instance)) then
    error ("setInputSocket: Request for too great a socket number of" 
	   ^ (string_of_int destSock)) else
    if (destSock < 0) then error "Request for negative socket number"
    else insert dict (dest,new_instance);;
  
  
let ancestors dict g : circuitDictionary -> gateId -> [gateId] = 
  g::(List.flatten (List.map (ancestors dict) (parents dict g)   ))

let isAncestor dict a b : circuitDictionary -> gateId -> gateId -> bool =
 List.mem b (ancestors dict a);;
 
let parents (gates,wires,_) g : circuitDictionary -> gateId -> [gateId] =
 let gate_instance = List.assoc g gates in
 map (function x -> (List.assoc x wires).sourceGate) gate_instance.inps
;;

let compile dict : circuitDictionary -> Microcode.quantumProgram =
 let (gates,_,_) = dict in
 if (not (verifyDictionary dict)) then error "Dictionary corrupt" else
 map (toInstruction_alt gates) (snafflesort dict (all_gateIds dict))
;;
  
let toInstruction gate_instance =
 {gate_type = gate_instance.nature, inputs = gate_instance.inps,
   outputs = gate_instances.outs, 
   gate = Gatedefs.gate_operation gate_instance.nature
 };;

let toInstruction_alt dict g = toInstruction (List.assoc g dict)

let rec snafflesort dictionary gateids = match gateids with
 [] -> []
| (q::qs)  =  
  let anc x y = isAncestor dictionary x y and 
      desc x y = isAncestor dictionary y x in
  let ancestors = Util.filter (anc) qs and
      descendants = Util.filter (desc) qs and
      neither = Util.filter (function x -> (not (anc x)) && (not (desc x))) qs
  in
    (snafflesort ancestors) @ [q] @ (snafflesort neither) @ 
  (snafflesort descendants);;

let rec nosort (gates,_,_) gateids =
  map (function q -> List.assoc q gates) gateids;;
   
let all_gateIds (gates,_,_) = fst (List.split gates);;
let all_wireIds (_,wires,_) = fst (List.split wires);;
  
verifyDictionary :: CircuitDictionary -> Bool
let verifyDictionary dict = 
  let gates = all_gateIds dict and wires = all_wiresIds dict
  in 
  List.for_all (verify_outputs dict) gates &&
  List.for_all (verify_inputs dict) gates &&
  List.for_all (verify_source dict) wires &&
  List.for_all (verify_dest dict) wires;;

let verify_outputs (gates,wires,_) g = 
  List.for_all (function s -> (List.assoc s wires).sourceGate = g)
    ((List.assoc g gates).outs);;

let verify_inputs (gates,wires,_) g = 
  List.for_all (function s -> (List.assoc s wires).destGate = g)
    ((List.assoc g gates).inps);;

let verify_source (gates,wires,_) w =
  let g = (List.assoc w wires).sourceGate in
  List.mem w  (List.assoc g gates).outs;;

let verify_dest (gates,wires,_) w =
 let g = (List.assoc w wires).destGate in
  List.mem w (List.assoc g gates).inps;;
