(*  Converted from Haskell to CaML (Sun Aug 16 13:05:14 EST 1998) 
\chapter{Miscellaneous useful functions}

\section{Preliminaries}

*)
module MiscFuncs where

open GregComplex
open Type
open Complex
(*

\section{Fourier Transform}
The \texttt{metric} value is only used when I get paranoid about 
numeric precision and want to check that the fourier transform is
doing an OK job.

*)
fourierTransform :: Integer -> Integer -> [(GregComplex,Integer)]
fourierTransform size row = answer
 where answer = [(raise_around size row column,column) |
                                 column <- [0..(size-1)]]
       metric = sum (map (\x -> (magnitude x)^2) (fst (unzip answer)))
(*

\texttt{raise\_around} picks on entry out of the following matrix \ldots 
\[
\frac{1}{\sqrt{N}}
\left(
\begin{array}{ccccc}
1 &          1        &             1      &             1  & \ldots \\
1 & e^{\left(\frac{2 \pi i}{N}\right)} & e^{{\left(\frac{2 \pi i}{N}\right)}^2} &  e^{{\left(\frac{2 \pi i}{N}\right)}^3} & \ldots \\
1 & e^{{\left(\frac{2 \pi i}{N}\right)}^2} & e^{{\left(\frac{2 \pi i}{N}\right)}^4} &  e^{{\left(\frac{2 \pi i}{N}\right)}^6} & \ldots \\
1 & e^{{\left(\frac{2 \pi i}{N}\right)}^3} & e^{{\left(\frac{2 \pi i}{N}\right)}^6} &  e^{{\left(\frac{2 \pi i}{N}\right)}^9} & \ldots \\
\vdots & \vdots & \vdots & \vdots & \ddots
\end{array}
\right)
\]

\begin{description}
\item[size] The array is square,  size gives number of rows and columns.  Size is abbreviated to \texttt{N} in the above matrix
\item[row] Which row to return
\item[column] Which column
\end{description}

*)
raise_around :: Integer -> Integer -> Integer -> GregComplex
raise_around size row column = 
   (cis (real_row * real_column* iteration_segment)) * one_on_root_N
  where iteration_segment = 2.0 * pi / (fromIntegral size)
        one_on_root_N = recip ((sqrt (fromIntegral size)) :+ 0.0)
        real_row = fromIntegral row
        real_column = fromIntegral column
        
(*
 
\section{ModuloPower}

\texttt{ModuloPower} calculates $x^a \quad{}\textsf{mod}\quad{} n$
*)
moduloPower :: (Integer,Integer,Integer) -> Integer
moduloPower (x,a,n)
 | a < 0 = error "moduloPower called with a < 0"
 | x < 0 = error "moduloPower called with x < 0"
 | n < 0 = error "moduloPower called with n < 0"
 | a == 0 = 1
 | a == 1 = x `mod` n
 | a `mod` 2 == 1  = (x * squared_recursive_part) `mod` n
 | otherwise = squared_recursive_part
 where recursive_part = moduloPower (x,(a `div` 2),n)
       squared_recursive_part = (recursive_part ^ 2) `mod` n
(**)
