(*  Converted from Haskell to CaML (Sun Aug 16 13:05:14 EST 1998) 
\chapter{Qfunc}

This module handles lots of things to do with quantum functions,  such
as evaluating them when required,  and compiling them, and such like.

\section{Preliminaries}

*)
module Qfunc where

open Type
open Microcode
open GateDefs
open RecognizeNumbers
open ReadDiagram
open IO
(*

\section{Evaluations}
*)
evaluateQfunc :: QuantumProgram -> [Type] -> [Double] -> [Type]
evaluateQfunc program arguments random_list = return_result
  where modified_program = insertArguments program arguments
        results = runQuantum modified_program random_list
        (result_positions,result_answers) = unzip results
        last_result = maximum result_positions
        return_result = [whichever_one_has_x_as_first_part x results |
                             x <- [0..last_result]]

(*

\section{Substitutions}
\label{Call site substitutions}

\textit{Source} instructions
can come with a string argument,   which will be of the form
\texttt{input:3}.  This ``3'' is a number (position) in a
argument list given to a cgol function.

Quantum functions get called just like classical functions, and
somehow we need to get these positional arguments replaced so that
they act more like initial values.  So we need to find every
\textit{Source} gate which is of the form mentioned, and replace it by
a \textit{Source} gate which is a source for the appropriate value in
the cgol call site argument list.

This code is not very clear,  and probably could do with a rewrite.

\subsection{insertArguments}

\textsf{insertArguments} is just a glorified map of replaceInstruction

*)
insertArguments :: QuantumProgram -> [Type] -> QuantumProgram
insertArguments [] _ = []
insertArguments (instruction:instructions) arguments
 = (replaced_instruction : insertArguments instructions arguments)
 where replaced_instruction = replaceInstruction instruction arguments
(*

\subsection{replaceInstruction}
\textsf{replaceInstruction} only cares about \textit{Source} instructions.
Everything else it just returns unchanged.  \textit{Source} instructions
get passed onto \textsf{bingo},  which does the magical change.

*)
replaceInstruction :: QuantumProgramInstruction -> [Type] -> QuantumProgramInstruction
replaceInstruction instruction arguments
 = case (gate_type instruction) of
     (Source (Str x)) -> bingo instruction x arguments
     _ -> instruction
(*

\subsection{bingo}

See above.  \textsf{bingo} does the replacement.

*)
bingo :: QuantumProgramInstruction -> String -> [Type] -> QuantumProgramInstruction
bingo instruction x arguments
   = if (RecognizeNumbers.is_an_input_argumentor x)  
          then (Instruction 
                       {gate_type = Source value,
                        gate = gate_operation (Source value),
                        inputs = inputs instruction,
                        outputs = outputs instruction
                        })
          else  instruction
        where which_argument = which_input_argumentor x
              value = arguments !! which_argument
(*

\subsection{Helpers}

*)        
whichever_one_has_x_as_first_part :: Integer -> [(Integer,Type)] -> Type
whichever_one_has_x_as_first_part x [] = Monad
whichever_one_has_x_as_first_part x ((y,z):ys)
 | x == y = z
 | otherwise = whichever_one_has_x_as_first_part x ys


(*
*)
