(*  Converted from Haskell to CaML (Sun Aug 16 13:05:14 EST 1998) 
\chapter{Qfunc}

This module handles lots of things to do with quantum functions,  such
as evaluating them when required,  and compiling them, and such like.


\section{Substitutions}
\label{Call site substitutions}

\textit{Source} instructions
can come with a string argument,   which will be of the form
\texttt{input:3}.  This ``3'' is a number (position) in a
argument list given to a cgol function.

Quantum functions get called just like classical functions, and
somehow we need to get these positional arguments replaced so that
they act more like initial values.  So we need to find every
\textit{Source} gate which is of the form mentioned, and replace it by
a \textit{Source} gate which is a source for the appropriate value in
the cgol call site argument list.

This code is not very clear,  and probably could do with a rewrite.
*)

(*

\subsection{bingo}

See below.  \textsf{bingo} does the replacement.

*)

let bingo instruction x arguments =
  if (Recognize_numbers.is_an_input_argumentor x)  
  then (                       
    let value = List.nth arguments (Recognize_numbers.which_input_argumentor x)
    in 
    { Microcode.gate_type = Gatedefs.Source value;
      Microcode.gate = Gatedefs.gate_operation (Gatedefs.Source value);
      Microcode.inputs = instruction.Microcode.inputs;
      Microcode.outputs = instruction.Microcode.outputs
    })
  else  instruction;;


(*

\subsection{replaceInstruction}
\textsf{replaceInstruction} only cares about \textit{Source} instructions.
Everything else it just returns unchanged.  \textit{Source} instructions
get passed onto \textsf{bingo},  which does the magical change.

*)

let replaceInstruction instruction arguments=
  match instruction.Microcode.gate_type with 
    (Gatedefs.Source (Qgol_type.Str x)) -> bingo instruction x arguments
  |  _ -> instruction;;


(*
\subsection{insertArguments}

\textsf{insertArguments} is just a glorified map of replaceInstruction

*)


let rec insertArguments program arguments = match program with
 [] -> []
| (instruction::instructions) -> 
 let replaced_instruction = replaceInstruction instruction arguments in
 (replaced_instruction :: ( insertArguments instructions arguments));;

(*

\subsection{Helpers}

*)  
let rec whichever_one_has_x_as_first_part list (x : int) = match list with
  [] -> Qgol_type.Monad
| ((y,z)::ys) when x = y -> z
| _ -> whichever_one_has_x_as_first_part (List.tl list) x


(*
\section{Evaluations}
*)



let evaluateQfunc program arguments  =
  let modified_program = insertArguments program arguments in
  let results = Microcode.runQuantum modified_program in
  let (result_positions,result_answers) = List.split results in
  let last_result = Util.maximum result_positions and
      first_result = Util.minimum result_positions 
  in  List.map (whichever_one_has_x_as_first_part results) 
    (Util.list_range first_result last_result);;
