(*  Converted from Haskell to CaML (Sun Aug 16 13:05:14 EST 1998) 
\chapter{ReadDiagram}

\section{readConnections}

This probably looks clumsy.  Way down in \textsf{readExecutable} I do
a \textsf{getContents} on the file handle we are reading.  Clean didn't
have any such magical concept.

\textsf{getContents} makes things much easier.  Now I don't have to return
a unique-strict file object -- I just pass around a list of lines in the 
file.  (Which means I don't have to return anything inessential).

I should probably rewrite \textsf{readConnectionsSource} as some kind
of \textsf{foldr} operation (that's all it is), since the compiler
might be able to optimise that better, and it makes for more compact
code.

*)

let readConnections s id dict =
 let my_words = Util.split_words s in
 let wrong_length = (List.length my_words) mod 2 = 1 in
 let rec readConnections_alt numbers destSock dictionary =
   match numbers with []->dictionary
   | (idstr::(sockstr::others)) ->
     readConnections_alt others (destSock+1) (
              GateDictionary.makeConnection dictionary (id,destSock) (int_of_string idstr,int_of_string sockstr))
   | (x::xs) -> (prerr_endline "Cannot happen" ; exit 1)
 in
 if (List.length my_words) = 0 then dict
 else if (s = "") then dict
 else if wrong_length then (prerr_endline ("What should I do with ``" ^s^"''?") ; exit 1)
 else
   readConnections_alt my_words 1 dict

let rec readConnectionsSource contents line_num dictionary = 
  match contents with
    [] -> dictionary
  | (line::lines) -> readConnectionsSource lines (line_num+1) (readConnections line line_num dictionary)



let rec readExecutableSource contents line_num =
  if line_num = -1 then readExecutableSource (List.tl contents) 0 else
  match contents with
    [] -> (prerr_endline "unexpected end of file" ; exit 1)
  | (line::lines) when line = "---gate.x gate.y gate.type gate.attribute---"
      -> readExecutableSource lines 1
  | (line::lines) when line = "---gate outputs---" -> (lines,[])
  | (line::lines) ->
      let (unhandled_lines,other_gates) = readExecutableSource lines (line_num + 1)
	  and
          (_::(_::(arg3::(arg4::_)))) = Util.split_words line in
      let my_gate = Gatedefs.strings2gate arg3 arg4 in
      (unhandled_lines,(my_gate,line_num)::other_gates)
;;


let readline channel = try Some (input_line channel) with End_of_file -> None;;
let rec readlines channel = match (readline channel) with
    (Some l) -> l::(readlines channel)
   | None -> [];;

          
let readExecutable channel = 
 let contents = readlines channel in
 let (remaining_lines,whatever) = (readExecutableSource contents (-1)) in
 let quantum_function = readConnectionsSource remaining_lines 1 (GateDictionary.makeDictionary whatever) in
    GateDictionary.compile quantum_function;;


(**)
