(* lots of useful functions that I like *)

let rec list_range min max =
  if (min >= max) then [min] else min:: (list_range (min+1) max);;

(* let rec lazy_list min = [min] @ lazy lazy_list (min+1) *)

let rec string_to_list str =
  let len = String.length str in
  if len = 0 then [] else
  let positions = list_range 0 (len-1) in
   List.map (function x -> str.[x]) positions;;

let rec char_list_to_string charlist = match charlist with
  [] -> ""
| (c::cs) -> (String.make 1 c) ^ (char_list_to_string cs);;
  

let rec string_join joinage str_list =
  match str_list with
   [] -> ""
  | [word] -> word
  | w::ws -> w ^ joinage ^ (string_join joinage ws);;


let pi = 3.14159265358979323846264338;;

let filter pred values = 
  let prepend_if value list = if (pred value) then value::list else list in
      List.fold_right prepend_if values [];;

let partition pred values = 
  let partition_if value (yes_list,no_list) = 
      if (pred value) then (value::yes_list, no_list) 
                      else (yes_list, value::no_list) 
   in
      List.fold_right partition_if values ([],[]);;

let rec until pred values = match values with
 [] -> []
| x::xs when pred x -> []
| x::xs -> x::(until pred xs);;

let rec splitWhen pred values = match values with
 [] -> ([],[])
| x::xs when pred x -> let (r1,r2) = splitWhen pred xs in (x::r1,r2)
| x::xs -> ([],x::xs);;

let rec takeWhile pred values = 
  fst (splitWhen (function x -> not (pred x)) values);;

let rec dropWhile pred values = 
  snd (splitWhen (function x -> pred x) values);;

let rec remove_repetitions list = match list with
  [] -> []
| [x] -> [x]
| (x::(y::zs)) when x = y -> remove_repetitions (y::zs)
| (x::xs) -> x::(remove_repetitions xs)
;;

let whitespace = " \t\r";;

let is_whitespace letter = List.mem letter (string_to_list whitespace);;

let string_lstrip str = 
 (char_list_to_string (dropWhile is_whitespace (string_to_list str)));;

(* terribly inefficient! *)
let string_rstrip str = 
 (char_list_to_string 
    (List.rev 
       (string_to_list 
	  (string_lstrip 
	     (char_list_to_string 
		(List.rev 
		   (string_to_list str)
		   )
		)
	     )
	  )
       )
    );;

let string_strip str = string_lstrip (string_rstrip str);;


let split_words sentence = 
  let rec break_up line = match line with
    [] -> [[]]
  | (x::y::zs) when (is_whitespace x) && (is_whitespace y) -> 
            (break_up (y::zs))
  | (x::line) when (is_whitespace x) -> []::(break_up line)
  | letter::line ->  let (this_word::remainder) = break_up line in
    (letter::this_word)::remainder 
  in
  List.map char_list_to_string (break_up (string_to_list (string_strip sentence)));;

let rec maximum list =
  match list with
    [] -> (prerr_endline "maximum of empty list" ; exit 1)
  | [x] -> x
  | x::xs -> let m = maximum xs in if x > m then x else m;;

let rec minimum list =
  match list with
    [] -> (prerr_endline "minimum of empty list" ; exit 1)
  | [x] -> x
  | x::xs -> let m = minimum xs in if x < m then x else m;;



(*  let len = String.length str in
  if len = 0 then [] else [ str.[0] ] @ string_to_list (String.sub str 1 (len-1)) ;;*)
